/*
Copyright 2010 Harteex Productions
License: Creative Commons Attribution 3.0 Unported
http://creativecommons.org/licenses/by/3.0/
*/
#include "drawing.h"

display* gameDisplay  = NULL;

char debugString[256];

sprite* gameTiles;

gfx_font* gameFont              = NULL;
gfx_font* gameFontYellow        = NULL;
gfx_font* gameFontGrey          = NULL;
gfx_font* gameFontBlack         = NULL;
gfx_texture* gameSplash         = NULL;
gfx_texture* menuBg             = NULL;
sprite* hudSmallKeyRed          = NULL;
sprite* hudSmallKeyBlue         = NULL;
sprite* hudSmallKeyGreen        = NULL;
sprite* hudSmallKeyYellow       = NULL;
sprite* hudFireBootsSmall       = NULL;
sprite* hudFlippersSmall        = NULL;
sprite* hudSkatesSmall          = NULL;
sprite* hudSuctionBootsSmall    = NULL;
sprite* checkmark               = NULL;
sprite* yellowArrow             = NULL;

char        hudTimeLeft[6];
char        hudLevelNo[6];
char        hudChipsLeft[6];
char        hudPasswordStr[20]; // "password: " + 10 (maxlength pw incl \0)

uint8_t  levelListEntriesVisible = 16;

inline uint32_t gfx_color_back(gfx_color inColor) {
		return ((inColor & 0xF800) << 8) | ((inColor & 0x07E0) << 5) | ((inColor & 0x001F) << 3);
}


// MAJOR DRAWING METHODS

void drawSplash()
{
	gfx_render_target_clear(COLOR_WHITE);
	gfx_tex_draw(((gameDisplay->width - gameSplash->width) >> 1), ((gameDisplay->height - gameSplash->height) >> 1), gameSplash);
}

void drawMenu(uint8_t index)
{
	gfx_tex_draw(0, 0, menuBg);
	
	uint8_t start_y = 115;
	uint8_t offset_y = 18;
	uint8_t start_x = 20;
	uint8_t offset_x = 18;
	uint8_t i;
	for (i = 0; i < menu_length; i++)
	{
		if (i == index)
		{
			drawMenuSelection(start_x + offset_x * i, start_y + offset_y * i, strlen(main_menu[i]));
			gfx_font_print(start_x + offset_x * i, start_y + offset_y * i, gameFontBlack, main_menu[i]);
		}
		else
		{
			gfx_font_print(start_x + offset_x * i, start_y + offset_y * i, gameFont, main_menu[i]);
		}
	}
}

void drawGame()
{
	gfx_render_target_clear(COLOR_BLACK);
	drawGrid();
	drawHud();
	switch (gameModeSub)
	{
		case GAME_MODE_GAME_SPLASH:
			drawLevelSplash();
			break;
		case GAME_MODE_GAME_PLAY:
			break;
		case GAME_MODE_GAME_WON:
			drawLevelWon();
			break;
		case GAME_MODE_GAME_GAMEOVER:
			drawLevelGameOver();
			break;
		case GAME_MODE_GAME_PASSWORD:
			gfx_rect_fill_draw(tileFieldOffsetX, tileFieldOffsetY, tilesVisibleWidth * TILESIZE, tilesVisibleHeight * TILESIZE, COLOR_BLACK);
			drawPasswordInput();
			break;
	}

	char tempStr[50];
	sprintf(tempStr, "%s (%u/%u)", exitConfirmMenu[4], numberOfSkippedLevels(), SKIPPABLE_LEVELS);
	if (exitConfirmDialog)
	{
		fadescreen(40);

		uint16_t startX = 70;
		uint16_t startY = 60;
		uint16_t offsetY = 15;
		drawMenuWindow(startX, startY, 135, "*** PAUSED ***");
		uint8_t i;
		for (i = 0; i < exitConfirmMenuLength; i++)
		{
			if (i == exitConfirmMenuSelectedItem)
			{
				if (i == 4)
				{
					drawMenuSelection(startX + 52, startY + 25 + offsetY * i, strlen(tempStr));
					gfx_font_print(startX + 52, startY + 25 + offsetY * i, gameFontBlack, tempStr);
				}
				else
				{
					drawMenuSelection(startX + 52, startY + 25 + offsetY * i, strlen(exitConfirmMenu[i]));
					gfx_font_print(startX + 52, startY + 25 + offsetY * i, gameFontBlack, exitConfirmMenu[i]);
				}
			}
			else if (i == 4)
			{
				if (isCurLevelSkippable())
					gfx_font_print(startX + 52, startY + 25 + offsetY * i, gameFont, tempStr);
				else
					gfx_font_print(startX + 52, startY + 25 + offsetY * i, gameFontGrey, tempStr);
			}
			else
			{
				gfx_font* tempFont = gameFont;
				if (cur_level->number >= numberOfLevels && i == 2)
					tempFont = gameFontGrey;
				if (cur_level->number <= 1 && i == 3)
					tempFont = gameFontGrey;
				gfx_font_print(startX + 52, startY + 25 + offsetY * i, tempFont, exitConfirmMenu[i]);
			}
		}
	}
}

void drawError(char* errorMsg, char* errorMsg2)
{
	gfx_render_target_clear(COLOR_BLACK);
	if (errorMsg2[0] == '\0')
	{
		gfx_font_print_center(120, gameFont, errorMsg);
	}
	else
	{
		gfx_font_print_center(110, gameFont, errorMsg);
		gfx_font_print_center(130, gameFont, errorMsg2);
	}
}

void drawLevelList(uint16_t offset, uint16_t selected)
{
	fadescreen(40);

	uint16_t startX = 65;
	uint16_t startY = 30;
	uint8_t i;
	uint16_t levelIndex;
	char out[256];
	bool lastCompleted = false;

	drawMenuWindow(startX, startY, 187, "Level Select");

	for (i = 0; i < levelListEntriesVisible; i++)
	{
		levelIndex = i + offset;
		if (levelIndex == 0)
			lastCompleted = true;
		gfx_font* font;

		if (levels[levelIndex].completed)
		{
			sprite_draw(checkmark, startX + 10, startY + 20 + i * 10, 0);
		}
		else if (levels[levelIndex].skipped)
		{
			sprite_draw(yellowArrow, startX + 10, startY + 20 + i * 10, 0);
		}

		if (levels[levelIndex].completed || lastCompleted)
		{
			sprintf(out, "%3i %s\n", levelIndex + 1, levels[levelIndex].name);
			font = gameFont;
		}
		else
		{
			sprintf(out, "%3i %s\n", levelIndex + 1, "locked");
			font = gameFontGrey;
		}

		lastCompleted = levels[levelIndex].completed || levels[levelIndex].skipped;

		if (levelIndex == selected)
		{
			font = gameFontYellow;
		}

		gfx_font_print(startX + 25, startY + 20 + i * 10, font, out);
	}
}

void drawVictoryAnimation()
{
	gfx_render_target_clear(COLOR_BLACK);

	gfx_font_print_center(10, gameFont, "VICTORY");
	gfx_font_print_center(40, gameFont, "You've completed the challenges!");
	gfx_font_print_center(70, gameFont, "Melinda herself offers Chip membership");
	gfx_font_print_center(80, gameFont, "in the Bit Busters computer club.");
	gfx_font_print_center(110, gameFont, "Chip's dream has finally come true!");
	uint8_t i, j;
	for (i = 0; i < NMBR_OF_ROCKETS; i++)
	{
		if (rockets[i].active && rockets[i].delay == 0)
		{
			gfx_rect_fill_draw(rockets[i].x / 100, rockets[i].y / 100, 2, 2, COLOR_WHITE);
		}

		// draw sparks
		for (j = 0; j < 10; j++)
		{
			if (rockets[i].sparks[j].ttl > 0)
			{
				gfx_rect_fill_draw((rockets[i].sparks[j].x / 100) - 1, (rockets[i].sparks[j].y / 100) - 1, 3, 3, rockets[i].sparks[j].colorsec);
				gfx_rect_fill_draw((rockets[i].sparks[j].x / 100), (rockets[i].sparks[j].y / 100), 1, 1, rockets[i].sparks[j].colorpri);
			}
		}
	}

	if (victoryFade < 100)
		fadescreen(victoryFade);
}

void drawInstructions()
{
	fadescreen(40);

	uint16_t startX = 35;
	uint16_t startY = 30;

	drawMenuWindow(startX, startY, 187, "Instructions");
	gfx_font_print(startX + 10, startY + 20, gameFont, "Help Chip win the respect of Melinda,\nto join the Bit Busters club. For this,\nChip must complete a number of\nchallenges set up by Melinda.");
	gfx_font_print(startX + 10, startY + 75, gameFont, "Chip must collect all the required\nmicro chips to be able to pass the\nsocket, and get to the goal.\nFor more information, read the hints\ningame by stepping on the hint tiles.");
	gfx_font_print(startX + 10, startY + 140, gameFont, "Micro Chip  Socket  Goal/Exit    Hint");
	uint8_t i = 0;
	sprite_draw(gameTiles, startX + 25 + 61 * i++, startY + 155, TILE_CHIP);
	sprite_draw(gameTiles, startX + 25 + 61 * i++, startY + 155, TILE_SOCKET);
	sprite_draw(gameTiles, startX + 25 + 61 * i++, startY + 155, TILE_EXIT);
	sprite_draw(gameTiles, startX + 25 + 61 * i++, startY + 155, TILE_HINT);
}

void drawControls()
{
	fadescreen(40);

	uint16_t startX = 40;
	uint16_t startY = 55;

#if COMPILE_TARGET == GEMEI_760P
	drawMenuWindow(startX, startY, 80, "Controls");
	gfx_font_print(startX + 10, startY + 20, gameFont, "DPAD                Move Player\nB                   In-game Menu\nTriangle            Restart Level\nX+Left/Right        Change Level\nX+Up/Down           Change Volume");
#else
	drawMenuWindow(startX, startY, 118, "Controls");
	gfx_font_print(startX + 10, startY + 20, gameFont, "DPAD                Move Player\nStart               In-game Menu\nB                   In-game Menu\nX                   Restart Level\nY+Left/Right        Change Level\nY+Up/Down           Change Volume\nR/L                 Change Volume\nStart+Select        Exit Game\nSelect (hold)       Take Screenshot");
#endif
}

void drawOptions(uint8_t curOption)
{
	fadescreen(40);

	uint16_t startX = 60;
	uint16_t startY = 45;
	uint16_t optionOffset = 100;

	drawMenuWindow(startX, startY, 150, "Options");

	if (curOption == 0)
	{
		drawMenuSelection(startX + 30, startY + 20, strlen("Sound"));
		gfx_font_print(startX + 30, startY + 20, gameFontBlack, "Sound");
	}
	else
	{
		gfx_font_print(startX + 30, startY + 20, gameFont, "Sound");
	}
	gfx_font_print(startX + 30, startY + 24, gameFont, "_____");
	gfx_font_print(startX + 30, startY + 37, configSound == CONFIG_SOUND_ON ? gameFontYellow : gameFont, "On");
	gfx_font_print(startX + 30, startY + 49, configSound == CONFIG_SOUND_REDUCED ? gameFontYellow : gameFont, "Reduced");
	gfx_font_print(startX + 30, startY + 61, configSound == CONFIG_SOUND_OFF ? gameFontYellow : gameFont, "Off");

	if (curOption == 1)
	{
		drawMenuSelection(startX + 30 + optionOffset, startY + 20, strlen("Music"));
		gfx_font_print(startX + 30 + optionOffset, startY + 20, gameFontBlack, "Music");
	}
	else
	{
		gfx_font_print(startX + 30 + optionOffset, startY + 20, gameFont, "Music");
	}
	gfx_font_print(startX + 30 + optionOffset, startY + 24, gameFont, "_____");
	gfx_font_print(startX + 30 + optionOffset, startY + 37, configMusic == CONFIG_MUSIC_ON ? gameFontYellow : gameFont, "On");
	gfx_font_print(startX + 30 + optionOffset, startY + 49, configMusic == CONFIG_MUSIC_OFF ? gameFontYellow : gameFont, "Off");

	// Explanation of current option at the bottom
	if (curOption == 0)
		gfx_font_print(startX + 30, startY + 100, gameFontGrey, "Help:\nOn - All Sounds On\nReduced - Some Sounds On\nOff - All Sounds Off");
	else
		gfx_font_print(startX + 30, startY + 100, gameFontGrey, "Help:\nOn - Music On\nOff - Music Off");
}

// curTime should start on zero
void _fadeText(int32_t curTime, int32_t duration, uint16_t x, uint16_t y, char* text)
{
	if (curTime > duration)
		curTime = duration;

	int32_t percent = (curTime * 100) / duration;

	gfx_font_print(x, y, gameFont, text);
	fadescreen(percent);
}

void _flyText(int32_t curTime, int32_t duration, uint16_t startX, uint16_t startY, char* text)
{
	if (curTime > duration)
		curTime = duration;

	uint16_t distance = 320 - startX;
	uint16_t x = 320 - ((curTime * distance) / duration);

	gfx_font_print(x, startY, gameFont, text);
}

void drawCredits()
{
	gfx_render_target_clear(COLOR_BLACK);
	
	uint32_t tempCounter = creditsAnimCounter;
	
	uint16_t startX = 40;
	uint16_t startY = 100;
	uint16_t introFadeDuration = 30;
	uint16_t introDuration = 60;
	uint16_t nameItemDuration = 30;

	uint8_t curItem = 0;
	if (tempCounter >= introDuration)
	{
		tempCounter -= introDuration;
		curItem++;

		// Determine current item to display
		uint32_t i;
		uint32_t itemTime;
		for (i = 0; i < creditsSize; i++)
		{
			itemTime = 100 + nameItemDuration * credits[i].nmbrOfNames;
			if (tempCounter < itemTime)
			{
				break;
			}

			curItem++;
			tempCounter -= itemTime;
		}
	}

	// Faded part
	if (curItem == 0)
	{
		if (tempCounter < introFadeDuration)
		{
			gfx_font_print_center(110, gameFont, "CREDITS");

			int32_t percent = (tempCounter * 100) / introFadeDuration;
			fadescreen(percent);
		}
		else
		{
			uint16_t duration = introDuration - introFadeDuration;
			uint16_t distance = 110 - 20;
			uint16_t y = 110 - (((tempCounter - introFadeDuration) * distance) / duration);
			gfx_font_print_center(y, gameFont, "CREDITS");
		}
	}
	else if (curItem <= creditsSize)
	{
		if (tempCounter >= 20)
		{
			uint32_t tempCounter2 = tempCounter - 20;

			uint32_t i;
			for (i = 0; i < credits[curItem - 1].nmbrOfNames; i++)
			{
				if (tempCounter2 >= nameItemDuration)
				{
					tempCounter2 -= nameItemDuration;
				}
				else
				{
					_fadeText(tempCounter2, 10, startX + 10, startY + 20 + (15 * i), credits[curItem - 1].names[i]);
					break;
				}
			}

			// After fade
			gfx_font_print(startX, startY, gameFont, credits[curItem - 1].header);

			uint32_t j;
			for (j = 0; j < i; j++)
			{
				gfx_font_print(startX + 10, startY + 20 + (15 * j), gameFont, credits[curItem - 1].names[j]);
			}
		}
		else
		{
			// Fly in header text
			_flyText(tempCounter, 15, startX, startY, credits[curItem - 1].header);
		}

		gfx_font_print_center(20, gameFont, "CREDITS");
	}
	else
	{
		uint16_t outroDuration = 170;
		uint16_t outroFadeStart = 50;
		if (tempCounter < outroDuration)
		{
			gfx_font_print_center(20, gameFont, "CREDITS");

			gfx_font_print(startX, startY, gameFont, credits[curItem - 2].header);

			uint32_t i;
			for (i = 0; i < credits[curItem - 2].nmbrOfNames; i++)
			{
				gfx_font_print(startX + 10, startY + 20 + (15 * i), gameFont, credits[curItem - 2].names[i]);
			}

			if (tempCounter >= outroFadeStart)
			{
				int32_t percent = ((tempCounter - outroFadeStart) * 100) / (outroDuration - outroFadeStart);
				percent -= 100;
				percent = abs(percent);
				fadescreen(percent);
			}
		}
	}
}

void drawExitScreen(uint8_t n)
{
	gfx_tex_draw(0, 0, menuBg);

	uint16_t x = 130;
	uint16_t y = 220;

	if (n == EXIT_SAVING)
	{
		gfx_font_print(x, y, gameFont, "Saving...");
	}
	else if (n == EXIT_EXITING)
	{
		gfx_font_print(x, y, gameFont, "Exiting...");
	}
}

// HELPER DRAWING METHODS

void fadescreen(int8_t percent)
{
	if (percent >= 100)
		return;
	if (percent < 0)
		percent = 0;

	uint16_t* tempTexPtr = gfx_render_target->address;
	uint32_t i;
	uint32_t totpx;
	uint16_t red;
	uint16_t green;
	uint16_t blue;

	totpx = gfx_render_target->width * gfx_render_target->height;
	for (i = 0; i < totpx; i++)
	{
		if (tempTexPtr[i] == 0)
			continue;

		red = (tempTexPtr[i] & 0xF800);
		green = (tempTexPtr[i] & 0x07E0);
		blue = (tempTexPtr[i] & 0x001F);

		red = (red * percent) / 100;
		green = (green * percent) / 100;
		blue = (blue * percent) / 100;

		red = red & 0xF800;
		green = green & 0x07E0;
		blue = blue & 0x001F;
		
		tempTexPtr[i] = (red | green | blue);
	}
}

void drawPasswordInput()
{
	uint16_t startX = 30;
	uint16_t startY = 80;
	uint8_t spaceX = 15;
	uint8_t spaceY = 15;

	gfx_font_print(startX, 30, gameFont, "ENTER PASSWORD:");
	gfx_font_print(startX, 50, gameFont, pwText);
	gfx_font_print(startX, 52, gameFont, "____");

	if (pwWrong > 0)
		gfx_font_print(startX, 160, gameFont, "Wrong password, try again...");

	uint8_t i;
	uint8_t x = 0;
	uint8_t y = 0;
	for (i = 0; i < 26; i++)
	{
		x = i % pwLetterWidth;
		y = i / pwLetterWidth;
		if (i == pwSelected)
		{
			drawMenuSelection(x * spaceX + startX, y * spaceY + startY, 1);
			gfx_font_print_char(x * spaceX + startX, y * spaceY + startY, gameFontBlack, 'A' + i);
		}
		else
		{
			gfx_font_print_char(x * spaceX + startX, y * spaceY + startY, gameFont, 'A' + i);
		}
	}

	uint16_t tempX = startX + (pwLetterWidth - 1) * spaceX + 6 - (6 * 3);
	if (pwSelected == 26)
	{
		drawMenuSelection(tempX, 3 * spaceY + startY, 3);
		gfx_font_print(tempX, 3 * spaceY + startY, gameFontBlack, "End");
	}
	else
	{
		gfx_font_print(tempX, 3 * spaceY + startY, gameFont, "End");
	}
}

void drawHud()
{
	if (getTile(playerX, playerY, 2) == TILE_HINT)
	{
		gfx_font_print(235, 20, gameFont, cur_level->hint);
		return;
	}

	uint16_t startX = 244;

	gfx_font_print(startX, 20, gameFont, "LEVEL");
	gfx_font_print(startX, 40, gameFont, hudLevelNo);
	
	gfx_font_print(startX, 75, gameFont, "TIME LEFT");
	if (gameTimeLeft <= 15)
		gfx_font_print(startX, 95, gameFontYellow, hudTimeLeft);
	else
		gfx_font_print(startX, 95, gameFont, hudTimeLeft);

	gfx_font_print(startX, 130, gameFont, "CHIPS LEFT");
	if (gameChipsLeft == 0)
		gfx_font_print(startX, 150, gameFontYellow, hudChipsLeft);
	else
		gfx_font_print(startX, 150, gameFont, hudChipsLeft);

	uint16_t fieldStartX = startX;
	uint16_t fieldStartY = 185;
	uint8_t x;
	uint8_t y;
	for (y = 0; y < 2; y++)
	{
		for (x = 0; x < 4; x++)
		{
			uint16_t xPx = fieldStartX + 16 * x;
			uint16_t yPx = fieldStartY + 16 * y;
			uint8_t size = 16;

			gfx_rect_fill_draw(xPx, yPx, size, size, gfx_color_rgb(0x60, 0x60, 0x60));
			gfx_line_draw(xPx, yPx, xPx + size - 1, yPx, gfx_color_rgb(0x90, 0x90, 0x90));
			gfx_line_draw(xPx, yPx, xPx, yPx + size, gfx_color_rgb(0x90, 0x90, 0x90));
			gfx_line_draw(xPx, yPx + size - 1, xPx + size - 1, yPx + size - 1, gfx_color_rgb(0x30, 0x30, 0x30));
			gfx_line_draw(xPx + size - 1, yPx, xPx + size - 1, yPx + size, gfx_color_rgb(0x30, 0x30, 0x30));

			/*gfx_rect_fill_draw(xPx, yPx, size, size, gfx_color_rgb(0x30, 0x30, 0x30));
			gfx_line_draw(xPx, yPx, xPx + size - 1, yPx, gfx_color_rgb(0x60, 0x60, 0x60));
			gfx_line_draw(xPx, yPx, xPx, yPx + size, gfx_color_rgb(0x60, 0x60, 0x60));
			gfx_line_draw(xPx, yPx + size - 1, xPx + size - 1, yPx + size - 1, gfx_color_rgb(0x10, 0x10, 0x10));
			gfx_line_draw(xPx + size - 1, yPx, xPx + size - 1, yPx + size, gfx_color_rgb(0x10, 0x10, 0x10));*/
		}
	}

	if (haveKeysRed > 0)
	{
		sprite_draw(hudSmallKeyRed, fieldStartX, fieldStartY, 0);
	}
	if (haveKeysBlue > 0)
	{
		sprite_draw(hudSmallKeyBlue, fieldStartX + 16, fieldStartY, 0);
	}
	if (haveKeysYellow > 0)
	{
		sprite_draw(hudSmallKeyYellow, fieldStartX + 32, fieldStartY, 0);
	}
	if (haveKeysGreen > 0)
	{
		sprite_draw(hudSmallKeyGreen, fieldStartX + 48, fieldStartY, 0);
	}
	if (haveIceSkates)
	{
		sprite_draw(hudSkatesSmall, fieldStartX, fieldStartY + 16, 0);
	}
	if (haveSuctionBoots)
	{
		sprite_draw(hudSuctionBootsSmall, fieldStartX + 16, fieldStartY + 16, 0);
	}
	if (haveFireBoots)
	{
		sprite_draw(hudFireBootsSmall, fieldStartX + 32, fieldStartY + 16, 0);
	}
	if (haveFlippers)
	{
		sprite_draw(hudFlippersSmall, fieldStartX + 48, fieldStartY + 16, 0);
	}
}

void drawGrid()
{
	int16_t x;
	int16_t y;
	for (y = 0; y < LEVELSIZE; y++)
	{
		for (x = 0; x < LEVELSIZE; x++)
		{
			int16_t newX = x + tileOffsetX;
			int16_t newY = y + tileOffsetY;

			if (newX >= 0 && newX < tilesVisibleWidth && newY >= 0 && newY < tilesVisibleHeight)
			{
				drawTile(x, y, tilesLayer1[x + y * LEVELSIZE]);
			}
		}
	}
}

void drawTile(uint16_t x, uint16_t y, uint16_t tile)
{
	uint16_t xPx = (x * TILESIZE) + tileFieldOffsetX + tileOffsetX * TILESIZE;
	uint16_t yPx = (y * TILESIZE) + tileFieldOffsetY + tileOffsetY * TILESIZE;

	// If tile is transparent, draw layer 2 first
	if (tile >= 0x3C && tile <= 0x6F)
	{
		// Draw layer 2
		uint8_t tile2 = getTile(x, y, 2);

		// If a tile on layer 2 happens to be transparent, draw the floor first, then draw transparent
		if  (tile2 >= 0x3C && tile2 <= 0x6F)
		{
			sprite_draw(gameTiles, xPx, yPx, TILE_FLOOR);
			sprite_draw(gameTiles, xPx, yPx, tile2);
		}
		else
		{
			sprite_draw(gameTiles, xPx, yPx, tile2);
		}

		// Draw layer 1
		sprite_draw(gameTiles, xPx, yPx, tile);
	}
	else // Tile is not transparent
	{
		sprite_draw(gameTiles, xPx, yPx, tile);
	}
}

void drawLevelSplash()
{
	uint16_t bannerWidth = 200;
	uint16_t bannerHeight = 50;
	uint16_t x = (tilesVisibleWidth * TILESIZE - bannerWidth) / 2 + tileFieldOffsetX;
	uint16_t y = 170;
	gfx_rect_fill_draw(x, y, bannerWidth, bannerHeight, gfx_color_rgb(0x00, 0x00, 0x00));
	gfx_font_print(x + 20, y + 10, gameFont, cur_level->name);
	gfx_font_print(x + 20, y + 30, gameFont, hudPasswordStr);
}

void drawLevelWon()
{
	uint16_t bannerWidth = 200;
	uint16_t bannerHeight = 90;
	uint16_t x = (tilesVisibleWidth * TILESIZE - bannerWidth) / 2 + tileFieldOffsetX;
	uint16_t y = 135;
	gfx_rect_fill_draw(x, y, bannerWidth, bannerHeight, gfx_color_rgb(0x00, 0x00, 0x00));
	gfx_font_print(x + 20, y + 10, gameFont, "LEVEL COMPLETE!");
	gfx_font_print(x + 20, y + 70, gameFont, "Press A for the next level");
}

void drawLevelGameOver()
{
	uint16_t bannerWidth = 200;
	uint16_t bannerHeight = 90;
	uint16_t x = (tilesVisibleWidth * TILESIZE - bannerWidth) / 2 + tileFieldOffsetX;
	uint16_t y = 135;
	gfx_rect_fill_draw(x, y, bannerWidth, bannerHeight, gfx_color_rgb(0x00, 0x00, 0x00));
	gfx_font_print(x + 20, y + 10, gameFont, "GAME OVER");
	gfx_font_print(x + 20, y + 30, gameFont, gameOverReason);
	gfx_font_print(x + 20, y + 70, gameFont, "Press A to restart level");
}

void drawVolumeIndicator(uint16_t cur, uint16_t max)
{
	uint16_t dotWidth = 12;

	uint16_t startX = (gameDisplay->width - (VOLUME_NMBR_OF_DOTS * dotWidth)) >> 1;

	uint16_t maxActiveDot = (cur * VOLUME_NMBR_OF_DOTS) / max;

	uint16_t x;
	uint16_t y = 200;
	uint16_t i;
	gfx_color tempCol = gfx_color_rgb(0x45, 0x45, 0x45);

	gfx_rect_fill_draw(startX - 6, y - 6, (VOLUME_NMBR_OF_DOTS * dotWidth) - 1 + 12, 10 + 12, tempCol);
	gfx_rect_draw(startX - 6, y - 6, (VOLUME_NMBR_OF_DOTS * dotWidth) - 1 + 12, 10 + 12, COLOR_BLACK);

	for (i = 0; i < VOLUME_NMBR_OF_DOTS; i++)
	{
		x = startX + i * dotWidth;

		if (i < maxActiveDot)
		{
			gfx_rect_fill_draw(x + 1, y + 1, 10, 10, COLOR_BLACK);
			gfx_rect_fill_draw(x, y, 10, 10, COLOR_WHITE);
		}
		else
		{
			gfx_rect_fill_draw(x, y, 10, 10, COLOR_BLACK);
		}
	}
}

void drawMenuSelection(uint16_t x, uint16_t y, uint16_t length)
{
	uint16_t width = length * 6 + 7;
	uint16_t height = 10;
	x -= 4;
	y -= 1;

	// Draw left side
	gfx_rect_fill_draw(x, y + 2, 1, height - 4, COLOR_WHITE);
	gfx_rect_fill_draw(x + 1, y + 1, 2, height - 2, COLOR_WHITE);

	// Draw right side
	gfx_rect_fill_draw(x + width - 1, y + 2, 1, height - 4, COLOR_WHITE);
	gfx_rect_fill_draw(x + width - 3, y + 1, 2, height - 2, COLOR_WHITE);

	// Draw middle
	gfx_rect_fill_draw(x + 3, y, width - 6, height, COLOR_WHITE);
}

void drawMenuWindow(uint16_t startX, uint16_t startY, uint16_t height, char* title)
{
	gfx_color headerColor = 0x4b99;
	gfx_rect_fill_draw(startX, startY, 320 - startX * 2, height, COLOR_BLACK);
	gfx_rect_draw(startX, startY, 320 - startX * 2, height, headerColor);
	gfx_rect_fill_draw(startX, startY, 320 - startX * 2, 15, headerColor);
	gfx_font_print_center(startY + 4, gameFont, title);
}
